package org.commonjava.tensor.web.base.rest;

import static javax.ws.rs.core.Response.Status.BAD_REQUEST;

import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Response;

import org.apache.maven.graph.common.DependencyScope;
import org.apache.maven.graph.common.ref.ProjectVersionRef;
import org.apache.maven.graph.effective.filter.DependencyFilter;
import org.apache.maven.graph.effective.filter.DependencyOnlyFilter;
import org.apache.maven.graph.effective.filter.ExtensionFilter;
import org.apache.maven.graph.effective.filter.ExtensionOnlyFilter;
import org.apache.maven.graph.effective.filter.OrFilter;
import org.apache.maven.graph.effective.filter.ParentFilter;
import org.apache.maven.graph.effective.filter.PluginOnlyFilter;
import org.apache.maven.graph.effective.filter.PluginRuntimeFilter;
import org.apache.maven.graph.effective.filter.ProjectRelationshipFilter;
import org.commonjava.tensor.agg.AggregationOptions;
import org.commonjava.tensor.agg.DefaultAggregatorOptions;
import org.commonjava.util.logging.Logger;

public final class RequestUtils
{

    private static final Logger LOGGER = new Logger( RequestUtils.class );

    private RequestUtils()
    {
    }

    public static ProjectVersionRef parseGAV( final String src )
    {
        String gav = src;
        if ( gav.startsWith( "/" ) )
        {
            if ( gav.length() > 1 )
            {
                gav = gav.substring( 1 );
            }
            else
            {
                return null;
            }
        }

        if ( gav.trim()
                .length() < 1 )
        {
            return null;
        }

        final String[] parts = gav.split( "\\/" );
        if ( parts.length != 3 )
        {
            throw new WebApplicationException(
                                               Response.status( BAD_REQUEST )
                                                       .entity( "Incomplete-graph sub-path should be of form 'some.group.id/artifact-id/version'" )
                                                       .build() );
        }

        final ProjectVersionRef ref = new ProjectVersionRef( parts[0], parts[1], parts[2] );
        return ref;
    }

    public static AggregationOptions createAggregationOptions( final HttpServletRequest request )
    {
        final DefaultAggregatorOptions options = new DefaultAggregatorOptions();
        options.setFilter( createGraphFilter( request ) );

        if ( getBooleanParamWithDefault( request, "disc", false ) )
        {
            options.setDiscoveryEnabled( true );

            options.setDiscoveryTimeoutMillis( getLongParamWithDefault( request, "to",
                                                                        options.getDiscoveryTimeoutMillis() ) );

            options.setProcessIncompleteSubgraphs( getBooleanParamWithDefault( request, "inc", true ) );
            options.setProcessIncompleteSubgraphs( getBooleanParamWithDefault( request, "var", true ) );
        }
        else
        {
            LOGGER.info( "discovery DISABLED for: %s", request.getRequestURL() );
        }

        LOGGER.info( "AGGREGATOR OPTIONS:\n\n%s\n\n", options );

        return options;
    }

    public static ProjectRelationshipFilter createGraphFilter( final HttpServletRequest request )
    {
        //        @QueryParam( "d" ) @DefaultValue( "true" ) final boolean dependencies,
        //        @QueryParam( "pa" ) @DefaultValue( "true" ) final boolean parents,
        //        @QueryParam( "pl" ) @DefaultValue( "false" ) final boolean plugins,
        //        @QueryParam( "e" ) @DefaultValue( "false" ) final boolean extensions,
        //        @QueryParam( "s" ) @DefaultValue( "runtime" ) final String scope 

        final List<ProjectRelationshipFilter> filters = new ArrayList<ProjectRelationshipFilter>();
        if ( getBooleanParamWithDefault( request, "d", true ) )
        {
            filters.add( new DependencyFilter( DependencyScope.getScope( getStringParamWithDefault( request, "s",
                                                                                                    "runtime" ) ) ) );
        }
        if ( getBooleanParamWithDefault( request, "pa", true ) )
        {
            filters.add( new ParentFilter( false ) );
        }
        if ( getBooleanParamWithDefault( request, "pl", false ) )
        {
            filters.add( new PluginRuntimeFilter() );
        }
        if ( getBooleanParamWithDefault( request, "e", false ) )
        {
            filters.add( new ExtensionFilter() );
        }

        final ProjectRelationshipFilter filter = new OrFilter( filters );
        LOGGER.info( "FILTER:\n\n%s\n\n", filter );

        return filter;
    }

    public static ProjectRelationshipFilter createDirectFilter( final HttpServletRequest request )
    {
        //        @QueryParam( "d" ) @DefaultValue( "true" ) final boolean dependencies,
        //        @QueryParam( "pa" ) @DefaultValue( "true" ) final boolean parents,
        //        @QueryParam( "pl" ) @DefaultValue( "false" ) final boolean plugins,
        //        @QueryParam( "e" ) @DefaultValue( "false" ) final boolean extensions,
        //        @QueryParam( "s" ) @DefaultValue( "runtime" ) final String scope 

        final List<ProjectRelationshipFilter> filters = new ArrayList<ProjectRelationshipFilter>();
        if ( getBooleanParamWithDefault( request, "d", true ) )
        {
            LOGGER.info( "ENABLE dependency filter." );
            filters.add( new DependencyOnlyFilter( DependencyScope.getScope( getStringParamWithDefault( request, "s",
                                                                                                        "runtime" ) ) ) );
        }
        if ( getBooleanParamWithDefault( request, "pa", true ) )
        {
            LOGGER.info( "ENABLE parent filte." );
            filters.add( new ParentFilter( false ) );
        }
        if ( getBooleanParamWithDefault( request, "pl", false ) )
        {
            LOGGER.info( "ENABLE plugin-only filter." );
            filters.add( new PluginOnlyFilter() );
        }
        if ( getBooleanParamWithDefault( request, "e", false ) )
        {
            LOGGER.info( "ENABLE extension-only filter." );
            filters.add( new ExtensionOnlyFilter() );
        }

        final ProjectRelationshipFilter filter = new OrFilter( filters );
        LOGGER.info( "FILTER:\n\n%s\n\n", filter );

        return filter;
    }

    public static String getStringParamWithDefault( final HttpServletRequest request, final String key, final String def )
    {
        String value = request.getParameter( key );
        final String val = value;
        if ( value == null || value.trim()
                                   .length() < 1 )
        {
            value = def;
        }

        LOGGER.info( "Value of key: %s is: %s. Returning string-param-with-default value: %s", key, value, val );
        return val;
    }

    public static long getLongParamWithDefault( final HttpServletRequest request, final String key, final long def )
    {
        final String value = request.getParameter( key );
        long val;
        if ( value == null || value.trim()
                                   .length() < 1 )
        {
            val = def;
        }
        else
        {
            val = Long.parseLong( value );
        }

        LOGGER.info( "Value of key: %s is: %s. Returning long-param-with-default value: %s", key, value, val );
        return val;
    }

    public static boolean getBooleanParamWithDefault( final HttpServletRequest request, final String key,
                                                      final boolean def )
    {
        final String value = request.getParameter( key );
        boolean val;
        if ( value == null || value.trim()
                                   .length() < 1 )
        {
            val = def;
        }
        else
        {
            val = Boolean.parseBoolean( value );
        }

        LOGGER.info( "Value of key: %s is: %s. Returning boolean-param-with-default value: %s", key, value, val );
        return val;
    }
}
